`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company: 
// Engineer: 
// 
// Create Date: 04.11.2022 21:07:37
// Design Name: 
// Module Name: tb_polmul
// Project Name: 
// Target Devices: 
// Tool Versions: 
// Description: 
// 
// Dependencies: 
// 
// Revision:
// Revision 0.01 - File Created
// Additional Comments:
// 
//////////////////////////////////////////////////////////////////////////////////


module tb_polmul( );

    reg clk;
    reg mult_rst;
    wire [63:0] mult_data64_in; 
         
    wire [9:0] mult_read_base_address, mult_write_base_address;
         
    wire mult_read_poly_op_sel;
    wire [9:0] mult_read_address, mult_write_address;
    wire mult_wea;
    wire mult_done;
    wire [63:0] mult_data64_out;

    wire bram_wea;
    wire [9:0] bram_addra;
    wire [9:0] bram_addrb;
    wire [63:0] bram_dina;
    wire [63:0] bram_doutb;
    
    reg tb_interrupt;
    reg tb_wea;
    reg [9:0] tb_addra, tb_addrb;
    reg [63:0] tb_dina;

    reg [63:0] test_data_a [255:0];
    reg [63:0] test_data_b [255:0];
    reg [8:0] i;
            
    polynomial_multiplication pmult(
         clk, 
         mult_rst,
         mult_read_poly_op_sel,
         mult_read_address, mult_write_address, 
         mult_wea,
         mult_data64_in, 
         mult_done,
         mult_data64_out
    );
    
    BRAM64_1024 BR (
         .clka(clk), // input clka --> Clock signal to port-A 
         .wea(bram_wea), // input [0 : 0] wea --> Write enable signal to port-A
         .addra(bram_addra), // input [9 : 0] addra --> Address to port-A
         .dina(bram_dina), // input [63 : 0] dina --> Input data to port-A
         .clkb(clk), // input clkb --> Clock for port-B
         .addrb(bram_addrb), // input [9 : 0] addrb --> Address to port-B
         .doutb(bram_doutb) // output [63 : 0] doutb --> Data output from port-B
    );

    assign mult_read_base_address = (mult_read_poly_op_sel) ? 10'd256 : 10'd0;
    assign mult_write_base_address = 10'd512;
    
    assign bram_wea = (tb_interrupt) ? tb_wea : mult_wea;
    assign bram_addra = (tb_interrupt) ? tb_addra : (mult_write_address+mult_write_base_address);
    assign bram_dina = (tb_interrupt) ? tb_dina : mult_data64_out;
    assign bram_addrb = (tb_interrupt) ? tb_addrb : (mult_read_address+mult_read_base_address);
    assign mult_data64_in = bram_doutb;


    initial begin
        clk = 0;
        mult_rst = 1;
        tb_interrupt = 1;


        tb_wea = 0;
        tb_addra = 0;
        tb_addrb = 0;
        tb_dina = 0;

        #100;

        $readmemh("C:/CryptoProcessor_2022_23/Verilog_src/poly_arithmetic/tb/poly_a.txt", test_data_a);
        $readmemh("C:/CryptoProcessor_2022_23/Verilog_src/poly_arithmetic/tb/poly_b.txt", test_data_b);        

        
        
        @(posedge clk);
        @(posedge clk);
        for(i=0; i<256; i=i+1)
        begin
            @(posedge clk);
            tb_wea = 1;
            tb_addra = i;
            tb_dina = test_data_a[i];
        end
       @(posedge clk);
       tb_wea = 0;
       tb_addra = 0;
       tb_dina = 0;
        

       @(posedge clk);
       @(posedge clk);
       for(i=0; i<256; i=i+1)
       begin
           @(posedge clk);
           tb_wea = 1;
           tb_addra = i+256;
           tb_dina = test_data_b[i];
       end
       @(posedge clk);
       tb_wea = 0;
       tb_addra = 0;
       tb_dina = 0;


        @(posedge clk);
        tb_interrupt = 0;
        
        @(posedge clk);
        mult_rst = 0;

        wait(mult_done);    

        
        @(posedge clk);
        mult_rst = 1;

        @(posedge clk);
        tb_interrupt = 1;

       @(posedge clk);
       @(posedge clk);
       for(i=0; i<256; i=i+1)
       begin
           tb_addrb = i+512;
           @(posedge clk);
       end
       @(posedge clk);

    end


    always #5 clk = ~clk;

endmodule
